;;; -*- auto-recompile: t -*-

;; This file is not part of GNU Emacs.
;; 
;; surmulot.el is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; surmulot.el is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:
;;-------------
;;             Surmulot integration 

;; last modified February 9, 2008
;; for Emacs 22
;;; --------------------------------------------------------------------------------

(require 'squeak)    
(require 'csound-x)

(require 'surmulot-widget)

;;; ==============================================================

(defun muo-connected-p ()
  (squeak-connected-somehow-p))

; (muo-get "'a,e,f' kmusic asBox")
(defun muo-get (object)
  (squeak-do (format "SurmulotManager emacsSend: (%s)" object)))

(defun muo-edit-string (encoded-string rid mode label)
  "Decode ENCODED-STRING and create a buffer to edit it
RID is a numeric handle allowing O to know what to do with the edited string
MODE is the major mode the edit buffer should be in
LABEL is the name for that buffer

This function is required by EmacsInteraction in O"
  (save-excursion
    (switch-to-buffer-other-frame
     (generate-new-buffer label))
    (erase-buffer)
    (insert (squeak-decode encoded-string))
    (goto-char (point-min))
    (condition-case nil
	(funcall mode)
      (error nil))
    (set (make-local-variable 'receiver-id) rid)
    (setq header-line-format
	  (list
	   (squeak-console-button-do 
	    " Accept " "accept the changes"
	    (squeak-do
	     (format "EmacsInteraction updateString: '%s' receiverId: %s " 
		     (squeak-encode-buffer)
		     receiver-id))
	    (set-buffer-modified-p nil))
	   " "
	   (squeak-console-button-do 
	    " Explore receiver " "open an ObjectExplorer in Squeak"
	    (squeak-do 
	     (format "(EmacsInteraction msDict at: %s) explore" receiver-id)))
	   " "
	   (squeak-console-button-do 
	    " Accept and exit " "accept the changes and kill this buffer"
	    (squeak-do 
             (format "EmacsInteraction updateString: '%s' receiverId: %s cleanUp: true" 
                     (squeak-encode-buffer)
		     receiver-id))
	    (let ((f (selected-frame)))
	      (when (and (kill-buffer (current-buffer))
			 (eq f (selected-frame)))
		(delete-frame))))
	   " "
	   (squeak-console-button-do 
	    " Exit " "kill this buffer and its frame"
	    (squeak-do 
	     (format "EmacsInteraction cleanUpReceiverId: %s" receiver-id))
	    (let ((f (selected-frame)))
	      (when (and (kill-buffer (current-buffer))
			 (eq f (selected-frame)))
		(delete-frame))))))
    (set-buffer-modified-p nil)))

;; (defun muo-get-buffer ()
;;    "send to O the appropriate MGraphElement for representing the current buffer"
;;    (interactive)
   
;; )

(defun muo-get-sco ()
  "send to O the appropriate MGraphElement for representing the current score"
  (interactive)
  (let ((filename (buffer-file-name)))
    (if (and filename (string-match "\\.sco$" filename))
        (muo-get (format "MGraphMusicFile newFromFile: '%s'" (squeak-filename filename)))
      (muo-get (format "MGraphCsoundScore newFromB64: '%s'" 
                          (if (cscsd-buffer-is-a-sco-p)
                              (squeak-encode-region (cscsd-sco-beginning) (cscsd-sco-end))
                            (squeak-encode-buffer)))))))

(defun muo-get-orc ()
  (interactive)
  "send to O the appropriate MGraphElement for representing the current orchestra"
  (let ((filename (buffer-file-name)))
    (if (and filename (string-match "\\.orc$" filename))
        (muo-get (format "MGraphCsoundComposer newFromFile: '%s'" (squeak-filename filename)))
      (muo-get (format "MGraphCsoundComposer newFromB64: '%s'" 
                          (if (cscsd-buffer-is-an-orc-p)
                              (squeak-encode-region (cscsd-orc-beginning) (cscsd-orc-end))
                            (squeak-encode-buffer)))))))

(defun muo-get-csd ()
  (interactive)
  "send to O the appropriate MGraphElement for representing the current csd"
  (let ((filename (buffer-file-name)))
    (if (and filename (string-match "\\.csd$" filename))
        (muo-get (format "MGraphMusicFile newFromFile: '%s'" (squeak-filename filename)))
      (muo-get (format "MGraphCsoundComposition newFromB64: '%s'" 
			  (squeak-encode-buffer))))))


;;; ============================ testing

(defun surmulot-assert (test result)
  (let ((val (squeak-eval test)))
    (or (string= val result) val)))	

(defun surmulot-tests ()
  (and
   (surmulot-assert "1+1" "2")
   (surmulot-assert "SurmulotManager askEmacsFor: '(+ 8 7)'" "15")
   ;; hairy... and fails !!
;   (surmulot-assert 
;    "SurmulotManager askEmacsFor: '(surmulot-assert \"2+2\" \"4\")'" "t")
))

;;; ============================ meta: surmulot system description

(defvar surmulot-directory (getenv "SURMULOTDIR")
  "root directory for the surmulot distribution")   

(defvar this-surmulot-version nil
  "holds the current surmulot version number")

(defvar this-surmulot-elisp-version nil
  "holds the current stef-elisp version number")

(defvar this-surmulot-muo-version nil
  "holds the current O version number")

(defvar this-surmulot-gm-version nil
  "holds the current GeoMaestro version number")

(defvar this-surmulot-published-files nil
  "holds the list of all files in the current surmulot release")

(defun surmulot-fetch-meta-info ()
  (when (boundp 'surmulot-directory) ;; this is defined in surmulot.emacs
    (load (expand-file-name "meta/surmulot-org.el" surmulot-directory) t)))

(surmulot-fetch-meta-info)

(defun surmulot-check-up ()
  (let ((modified))
    (dolist (fspec this-surmulot-published-files)
      (when (consp fspec)
        (let* ((file (car fspec))
               (fname (expand-file-name file surmulot-directory))
               (fmd5 (cdr fspec)))
          (unless (and (file-exists-p fname)
                       (string= fmd5 (with-temp-buffer
                                       (insert-file-contents fname)
                                       (md5 (current-buffer) nil nil 'raw-text))))
            (add-to-list 'modified file)))))
    (or modified t)))

;;TO DO:
;;       repair from archive (use 7za) !!


;;; ==============================================================

(defun surmulot-use-system-csound ()
  "experimental..."
  (interactive)
  (custom-set-default 'cscsd-csound-binary "csound")
  (cscsd-open-to-environment))

(defun surmulot-display-welcome-buffer ()
  (interactive)
  (let ((about (expand-file-name "emacs/about surmulot.org" surmulot-directory))
	(image-file (expand-file-name "emacs/surmulot.jpg" surmulot-directory)))
    (if (not (file-exists-p about))
	(info "(Surmulot)Top.")
      (find-file about)
      (when (file-exists-p image-file)
	(goto-char (point-min))
	(insert-image (create-image image-file)))
      (set-buffer-modified-p nil)
      (toggle-read-only))))

(defun surmulot-byte-compile-all ()
  "byte-compile all code for surmulot"
  (interactive)
  (dolist (dir '("emacs/site-lisp/stef-elisp" 
		 "emacs/site-lisp/stef-elisp/csound-x"))
    (dolist (file (directory-files 
		   (expand-file-name dir surmulot-directory) t "\\.el$"))
      (byte-compile-file file))))

(defun surmulot-go ()
  "Start the full surmulot system via emacs"
  (save-window-excursion
    (cd (file-name-directory (squeak-tcpip-launch-command)))
    (shell-command (format "%s &"
		     (shell-quote-argument (squeak-tcpip-launch-command)))))
  (save-window-excursion
    (cd (kk-tcpip-key-start-directory))
    (call-process (kk-tcpip-key-command) nil 0)))


;;========================= automatic deconnection upon killing

(defadvice save-buffers-kill-emacs (before deconnect-manager)
  "first deconnect emacs from its surmulot manager"
  (if (erman-connected-p)
      (erman-disconnect)))

(ad-activate 'save-buffers-kill-emacs)


;;========================= surmulot info

(when (require 'info nil t)
  (add-to-list 'Info-additional-directory-list (expand-file-name "emacs/" surmulot-directory)))


;;========================= surmulot menu

(defun surmulot-make-global-menu ()
  (easy-menu-define surmulot-menu global-map
    "surmulot menu"
    `("<surmulot>"
      ("Editing"
       ["cd to file directory" (cd (file-name-directory (buffer-file-name))) (buffer-file-name)]
       "--"
       ["get clipboard"    (progn (switch-to-buffer (get-buffer-create "*from clipboard*")) (clipboard-yank)) t]
       ["select all"                mark-whole-buffer t]
       ["untabify region"           untabify t]
       ["display search buffer"     qsheet-display-search-buffer t])
      "--"
      ["Build CSD from clipboard"          cscsd-import-from-clipboard t] 
      "--"
      ("Major modes"
       ["Csound/Orc" csound-orc-mode t]
       ["Csound/Sco" csound-sco-mode t]
       ["Keykit" keykit-mode t]
       ["Emacs lisp" (emacs-lisp-mode) t]
       ["Text" text-mode t])
      ("Minor modes"
       ["csound csd" (progn (text-mode)
			    (csound-csd-mode)
			    (cscsd-mmm)) t]
       ["eldoc" eldoc-mode :style toggle :selected  (condition-case nil eldoc-mode (error nil))])
      
      "--"
      ["Squeak console" (with-temp-buffer (require 'squeak) (display-squeak-tcpip-console)) t]
      ["Keykit console" (with-temp-buffer (require 'keykit-mode) (kk-tcpip)) t]
      ["System Shell"         shell t]
      "--"
      ["Connect to manager" erman-connect (not (erman-connected-p))]
      ["Disconnect from manager" erman-disconnect (erman-connected-p)]
      ["Start manager daemon" erman-start-daemon (and
						  (not (erman-daemon-waiting-p))
						  (not (erman-connected-p)))]
      ["Kill manager daemon" erman-cancel-daemon (erman-daemon-waiting-p)]
      "--"
      ["Customize paths" (customize-group 'spfa-paths) t]
      ["Customize Csound-x" (customize-group 'csound-x) t]
      ["Byte-compile all surmulot code" surmulot-byte-compile-all t]
      "--"
      ["Display welcome buffer" surmulot-display-welcome-buffer t]
      "--"
      ["Kill all processes" (mapcar 'kill-buffer (mapcar 'process-buffer (process-list))) (process-list)]
      ["Kill Emacs with no hook" (when (y-or-n-p "really kill emacs right now ?") (setq kill-emacs-hook nil) (kill-emacs)) t])))


;;; ==============================================================


(defcustom surmulot-timidity-cfg "$SURMULOTDIR/timidity/timidity.cfg"
  "Timidity configuration file for Surmulot"
  :type 'file
  :group 'spfa-paths)

(defun surmulot-timidity-cfg ()
  (expand-file-name (substitute-in-file-name surmulot-timidity-cfg)))

(defcustom surmulot-timidity-man "$SURMULOTDIR/timidity/man.html"
  "HTML manual for Timidity"
  :type 'file
  :group 'spfa-paths)

(defun surmulot-timidity-man ()
  (expand-file-name (substitute-in-file-name surmulot-timidity-man)))

(defcustom surmulot-keykit-man "$SURMULOTDIR/keykit/index.html"
  "HTML manual for Keykit"
  :type 'file
  :group 'spfa-paths)

(defun surmulot-keykit-man ()
  (expand-file-name (substitute-in-file-name surmulot-keykit-man)))



;;; --------------------------- End  -----------------------------

(provide 'surmulot)

;;; surmulot.el ends here







